/*
 * Routine conue par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#include <string.h>
#include <ctype.h>
#include <math.h>

#include "diffcycle_x86_64.h"
#include "lfx86_64.h"


unsigned long int createLi( Li* li , long int nb )
{
 long int i ;

 li->signe = 0 ;
 li->taille_mantisse = nb ;
 li->mantisse = (unsigned long int*) calloc( nb , sizeof(long int) ) ;
 for( i = 0 ; i < nb ; i++ ) li->mantisse[i] = 0 ;

 if (li->mantisse != NULL) return (nb * sizeof(long int) ) ;
 	else return 0 ;
}

void destroyLi( Li* li )
{
 free( li->mantisse ) ;
 li->taille_mantisse = -10 ;
 li->signe = -2 ;
}

long int indiceMax( long int* chiffre , Li* a )
{
 long int i , taille ;

 if ( a->signe == 0 )
	{
	 *chiffre = 0 ;
	 return 0 ;
	}
 taille = a->taille_mantisse ;
 for ( i = taille-1 ; i >= 0 ; i-- )
	if ( a->mantisse[i] != 0 ) break ;
 if ( i < 0 ) i = 0 ;
 *chiffre = a->mantisse[i] ;
 return i ;
}

void printLi( Li* li )
{
  long int i , n ;

 n = li->taille_mantisse ;
 printf("\n Precision = %ld" , n ) ;
 printf("\n Signe = %ld\n" , li->signe ) ;
 for( i = n - 1 ; i >= 0 ; i-- ) printf("%lu " , li->mantisse[i] ) ;
 printf("\n") ;
}

void copy( Li* nouveau , Li* li )
// Je suppose que li et nouveau ont mme taille.
{
 long int i , n ;

 nouveau->signe = li->signe ;
 nouveau->aux = li->aux ; 
 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ ) nouveau->mantisse[i] = li->mantisse[i] ;
}

void copyAndCreate( Li* nouveau , Li* li )
{
 long int  i , n ;

 nouveau->signe = li->signe ;
 n = li->taille_mantisse ;
 nouveau->taille_mantisse = n ;
 nouveau->mantisse = (long int*) calloc( n , sizeof(long int) ) ;
 for( i = 0 ; i < n ; i++ ) nouveau->mantisse[i] = li->mantisse[i] ;
}

void randomLi( Li* li , long positive )
// I suppose that lii is already created
// if positive == 1 then the integer is positive
// if positive == 0 then the the signis anywhere
{
 long int i , n ;
 unsigned long int t,t2 ;
 time_t rand_germe ;
 
 
 rand_germe = time(NULL) ;
 rand_germe *= rand_germe+2 ;
 rand_germe *= rand_germe ;
 rand_germe *= rand_germe ;
 srand((int) rand_germe ) ;

 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ )
	{
	 t  = (unsigned long int) rand() ;
	 t2 = (unsigned long int) rand() ;
	 t2 <<= 32 ;
	 t += t2 ;
	 li->mantisse[i] = t ;
	}

 n-- ;
 while ((n>=0) && ( li->mantisse[n] == 0 )) n-- ;
 if (n<0)
	{
	 li->signe = 0 ;
	}
	else
	{
	 if (positive == 1) li->signe = 1 ;
	 	else
		{
		 if (positive == 0) li->signe = (rand() % 2)*2-1 ;
			else li->signe = -2 ;
		}
	}
}

/*
void permutLi( Li* li2 , Li* li )
{
 long i , a ;

 a = li2->signe ;
 li2->signe = li->signe ;
 li->signe = a ;
 for( i = 0 ; i < li->taille_mantisse ; i++ )
 	{
	 a = li2->mantisse[i] ;
	 li2->mantisse[i] = li->mantisse[i] ;
	 li->mantisse[i] = a ;
	}
}
*/

//--------------------------------------------------------------------
// -------   Entier long en base 2^62      ---------------------------
//--------------------------------------------------------------------

/*
void add_R_2P_bin( Li* a , Li* b )
// a et b de mme taille et positif.
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 ) return ;
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
	 	a->mantisse[i] = b->mantisse[i] ;
         a->signe = b->signe ;
         return ;
        }
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] + b->mantisse[i] + carry ;
         carry = en >> DECALAGE32_BIN ;
         a->mantisse[i] = en & CACHE32_BIN ; // cache = 2^30 - 1
        }
}


void add_2P_bin( Li* result , Li* a , Li* b )
// Je suppose que result est dj cre et a la taille de a et b.
// a et b sont deux nombres positifs.
// Je ne gre pas le dbordement.
// result  <-  a + b
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = a->mantisse[i] ;
	 result->signe = a->signe ;
         return ;
        }
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = b->mantisse[i] ;
	 result->signe = b->signe ;
         return ;
        }

 result->signe = a->signe ;
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] + b->mantisse[i] + carry ;
         carry = en >> DECALAGE32_BIN ;
         result->mantisse[i] = en & CACHE32_BIN ;
	 //printf("%ld ** %ld %ld == %ld \n" , en , carry , result->Mantisse[i], CACHE32_BIN ) ;
        }
}

void sub_2P_bin( Li* result , Li* a , Li* b )
// Je suppose que a et b sont positifs.
// result  <-  a - b
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = a->mantisse[i] ;
         result->signe = a->signe ;
         return ;
        }
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = b->mantisse[i] ;
         result->signe = -(b->signe) ;
         return ;
        }
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] - b->mantisse[i] + carry ;
         carry = en >> 31 ;  // astuce pour rcuprer la retenue.
         result->mantisse[i] = en - carry * BASE32_BIN ;
	 //result->Mantisse[i] = en + (65536 >> ( 17 - (carry & 17) ) ) ;
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( result->mantisse[en] == 0 ) en++ ;
         result->mantisse[en] = BASE32_BIN - result->mantisse[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                result->mantisse[i] = CACHE32_BIN - result->mantisse[i] ;
         result->signe = -(a->signe) ;
         return ;
        }

 result->signe = a->signe ;
 for( i = 0 ; i < n ; i++ )
        if ( result->mantisse[i] != 0 ) break ;
 if ( i == n ) result->signe = 0 ;
}

void sub_R_2P_bin( Li* a , Li* b )
// Je suppose que a et b sont positifs.
// a  <-  a - b
{
 long i , en , carry , n ;

 if ( b->signe == 0 ) return ;
 n = a->taille_mantisse ;
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) a->mantisse[i] = b->mantisse[i] ;
         a->signe = -(b->signe) ;
         return ;
        }

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] - b->mantisse[i] + carry ;
         carry = en >> 31 ;
         a->mantisse[i] = en - carry * BASE32_BIN ;
	 //a->Mantisse[i] = en + (65536 >> ( 17 - (carry & 17) ) ) ;
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( a->mantisse[en]  == 0 ) en++ ;
         a->mantisse[en] = BASE32_BIN - a->mantisse[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                a->mantisse[i] = CACHE32_BIN - a->mantisse[i] ;
         a->signe = -(a->signe) ;
         return ;
        }

 for( i = 0 ; i < n ; i++ )
        if ( a->mantisse[i] != 0 ) break ;
 if ( i == n ) a->signe = 0 ;
}

void add_bin( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe >= 0 )
        {
         add_2P_bin( result , a , b ) ;
         return ;
        }
 sub_2P_bin( result , a , b ) ;
}

void add_R_bin( Li* a , Li* b )
// a <- a + b
{
 if ( a->signe * b->signe >= 0 )
        {
         add_R_2P_bin( a , b ) ;
         return ;
        }
 sub_R_2P_bin( a , b ) ;
}

void sub_bin( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe >= 0 )
        {
         sub_2P_bin( result , a , b ) ;
         return ;
        }
 add_2P_bin( result , a , b ) ;
}

void sub_R_bin( Li* a , Li* b )
// a <- a - b
{
 if ( a->signe * b->signe >= 0 )
        {
         sub_R_2P_bin( a , b ) ;
         return ;
        }
 add_R_2P_bin( a , b ) ;
}


void mulB_bin( Li* prod , Li* li , long m )
// multiplication par un chiffre.
// je ne gre pas le dbordement.
{
 long i , n ;
 unsigned long q  ;
 unsigned long aux ;

 n = li->taille_mantisse ;
 if (( li->signe == 0 ) || ( m == 0 ))
        {
         prod->signe = 0 ;
         for(i = 0 ; i < n ; i++ ) prod->mantisse[i]  = 0 ;
         return ;
        }
 prod->signe = li->signe ;
 q = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
	 aux = li->mantisse[i] ;
	 prod->mantisse[i] = mul_add_div_normalized_bin(&q,aux,m,q) ;
        }
}

long div_by_pow2_bin( Li* result , Li* li , long p )
// if faut que 0 < 2^p < BASE32_BIN.
// return the rest of the division.
{
 long i , a , r , q , n ;
 unsigned long tempon ;

 n = li->taille_mantisse ;
 if ( li->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = 0 ;
         return 0 ;
        }
 result->signe = li->signe ;
 a = 1 ;
 a <<= p ;

 q = (li->mantisse[n - 1]) >> p ;
 r = (li->mantisse[n - 1]) & (a-1) ;
 result->mantisse[n - 1] = q ;

 for( i = n - 2 ; i >= 0 ; --i )
        {
	 tempon = li->mantisse[i] ;
	 r = mul_add_div(&q,r,BASE32_BIN,tempon,a);
         result->mantisse[i] = q ;
        }
 for( i = 0 ; i < n ; ++i )
        if ( result->mantisse[i]  != 0 ) break ;
 if ( i == n ) result->signe = 0 ;
 return r ;
}

void mul_bin( Li* prod , Li* a , Li* b )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , n_prod ;
 unsigned long m , carry=0 ,tempon , aux ;

 n = a->taille_mantisse ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         m = b->mantisse[i] ;
         carry = 0 ;
         for( j = 0 ; j < n ; j++ )
                {
		 tempon = carry + prod->mantisse[i+j] ;
		 aux = a->mantisse[j] ;
		 prod->mantisse[i+j] = mul_add_div_normalized_bin(&carry,m,aux,tempon) ;
                }
         prod->mantisse[i+n] = carry ;
        }

 prod->signe = a->signe * b->signe ;
}

long mul_trunc_bin( Li* prod , Li* a , Li* b )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , np ;
 unsigned long m , carry=0 ,tempon , aux , pro = 0 ;

 n  = a->taille_mantisse ;
 np = n-1 ; // WARNING : PROBLEM WITH GCC-3.3.2 for n--  or  --n   ?????
 //printf("\n n = %ld  :: np = %ld " , n , np ) ;
 for( i = 0 ; i < n ; i++ ) prod->mantisse[i] = 0 ;

 for( i = 0 ; i < np ; i++ )
        {
         m = b->mantisse[i] ;
	 aux = a->mantisse[n-2-i] ;
	 pro = mul_add_div_normalized_bin(&carry,m,aux,pro ) ;
	 //printf("\n pro = %lu  :: m = %lu :: aux = %lu " , pro , m , aux ) ;
         for( j = np-i ; j < n ; j++ )
                {
		 tempon = carry + prod->mantisse[i+j-np] ;
		 aux = a->mantisse[j] ;
		 prod->mantisse[i+j-np] = mul_add_div_normalized_bin(&carry,m,aux,tempon) ;
                }
         prod->mantisse[i+1] = carry ;
        }

 m = b->mantisse[np] ;
 carry = 0 ;
 for( j = 0 ; j < n ; j++ )
	{
	 tempon = carry + prod->mantisse[j] ;
	 aux = a->mantisse[j] ;
	 prod->mantisse[j] = mul_add_div_normalized_bin(&carry,m,aux,tempon) ;
	}


 // teste du signe et de la nullit

 prod->signe = a->signe * b->signe ;
 for( i = 0 ; i < n ; i++ )
	if ( prod->mantisse[i] != 0 ) break ;
 if ( i == n ) prod->signe = 0 ;

 return (long) carry ;  // NO PROBLEM
}

void square_bin( Li* prod , Li* a )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , n_prod ;
 unsigned long p0 , p1 , p2 , p , tempon , tempon2 , aux , n2 ;

 n = a->taille_mantisse ;
 n2 = (n << 1)-3 ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;

 for( i = 0 ; i < n ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	 p0 += prod->mantisse[i] ;
	 prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	}
 for( i = n ; i < n2 ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = i-n+1 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 p0 += prod->mantisse[i] ;
	 prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	}

 // i = 2n-3
 tempon  = a->mantisse[n-2] ;
 tempon2 = tempon << 1 ;
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,0) ;
 p0 += prod->mantisse[n2] ;
 prod->mantisse[n2] = p0 & CACHE32_BIN ;
 aux += (prod->mantisse[n2+1] + ( p0 >> DECALAGE32_BIN ) ) ;
 prod->mantisse[n2+1] = aux & CACHE32_BIN ;
 //  i = 2n-2
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,0) ;
 prod->mantisse[n2+1] = ( p0 + prod->mantisse[n2+1] ) & CACHE32_BIN ;

 prod->signe = a->signe ;
}

long square_trunc_bin( Li* prod , Li* a )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , v ;
 unsigned long p0 , p1 , p2 , p , tempon , tempon2 , aux , n2 ;
 unsigned long t[5] = {0,0,0,0,0} ;

 n = a->taille_mantisse ;
 n2 = (n << 1)-3 ;
 v = 0 ;
 //n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n ; i++ ) prod->mantisse[i] = 0 ;


 for( i = n-3 ; i < n ; i++,v++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	 p0 += t[v] ; //p0 += prod->mantisse[i] ;
	 t[v] = p0 & CACHE32_BIN ; //prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (t[v+1] + ( p0 >> DECALAGE32_BIN ) ) ;//p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 t[v+1] = p1 & CACHE32_BIN ;  //prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( t[v+2] + ( p1 >> DECALAGE32_BIN ) ) ;//p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 t[v+2] = p2 & CACHE32_BIN ; //prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	}

 prod->mantisse[0] = t[2] ;
 prod->mantisse[1] = t[3] ;
 prod->mantisse[2] = t[4] ;

 for( i = n ; i < n2 ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = i-n+1 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 p0 += prod->mantisse[i-n+1] ;
	 prod->mantisse[i-n+1] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+2-n] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2-n] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+3-n] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+3-n] = p2 & CACHE32_BIN ;
	}

 // i = 2n-3
 tempon  = a->mantisse[n-2] ;
 tempon2 = tempon << 1 ;
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,0) ;
 p0 += prod->mantisse[n-2] ;
 prod->mantisse[n-2] = p0 & CACHE32_BIN ;
 //n2++ ;
 aux += (prod->mantisse[n-1] + ( p0 >> DECALAGE32_BIN ) ) ;
 prod->mantisse[n-1] = aux & CACHE32_BIN ;
 //  i = 2n-2
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,0) ;
 p0 += prod->mantisse[n-1] ;
 prod->mantisse[n-1] = p0 & CACHE32_BIN ;


 // tude du signe et test la nullit !
 prod->signe = 1 ;
 for( i = 0 ; i < n ; i++ )
	if ( prod->mantisse[i] != 0 ) break ;
 if ( i == n ) prod->signe = 0 ;

 return p0 >> DECALAGE32_BIN ;
}

void mul_by2_R_bin( Li* a )
{
 long i , carry , en , n ;

 n = a->taille_mantisse ;
 carry = 0 ;
 for(i = 0 ; i < n ; i++  )
        {
         en    = ( (a->mantisse[i]) << 1 ) + carry ;
         carry = en >> DECALAGE32_BIN ;
         a->mantisse[i]  = en & CACHE32_BIN ;
        }
}

double BigFloatToFloat( Li* bfBaseBin )
{
 long i , n ;
 double r ;

 n = bfBaseBin->taille_mantisse ;
 if ( bfBaseBin->signe == 0 ) return 0.0 ;
 r = bfBaseBin->mantisse[0] / 1073741824.0 ;
 for(i = 1 ; i < n - 1 ; i++ )
        r = ( r + (bfBaseBin->mantisse[i])) / 1073741824.0 ;
 r += bfBaseBin->mantisse[n - 1 ] ;
 return (r * (bfBaseBin->signe) ) ;
}

*/
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------

void createLii( Lii* li , long nb )
{
 long i ;

 li->signe = 0 ;
 li->indiceMax = -1 ;
 li->taille_mantisse = nb ;
 li->mantisse = (unsigned long int*) calloc( nb , sizeof(long int) ) ;
 for( i = 0 ; i < nb ; i++ ) li->mantisse[i] = 0 ;
}

void destroyLii( Lii* li )
{
 free( li->mantisse ) ;
 li->taille_mantisse = -10 ;
 li->signe = -1 ;
 li->indiceMax = -1 ;
}

void copyLii( Lii* nouveau , Lii* li )
// Je suppose que li et nouveau ont mme taille.
{
 long i , n ;

 nouveau->signe = li->signe ;
 nouveau->indiceMax = li->indiceMax ;
 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ ) nouveau->mantisse[i] = li->mantisse[i] ;
}

void randomLii( Lii* li , long int positive )
// I suppose that lii is already created
// if positive == 1 then the integer is positive
// if positive == 0 then the the signis anywhere
{
 long int i , n ;
 unsigned long int t,t2 ;
 //time_t rand_germe ;
 long int rand_germe ;
 
 
 aurora_x86_64( &rand_germe) ;
 //rand_germe = time(NULL) ;
 rand_germe *= rand_germe+2 ;
 rand_germe *= rand_germe ;
 srand((int) rand_germe ) ;

 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ )
	{
	 t  = (unsigned long int) rand() ;
	 t2 = (unsigned long int) rand() ;
	 t2 <<= 32 ;
	 t += t2 ;
	 li->mantisse[i] = t ;
	}

 n-- ;
 while ((n>=0) && ( li->mantisse[n] == 0 )) n-- ;
 if (n<0)
	{
	 li->signe = 0 ;
	 li->indiceMax = -2 ;
	}
	else
	{
	 if (positive == 1) li->signe = 1 ;
	 	else
		{
		 if (positive == 0) li->signe = (rand() % 2)*2-1 ;
			else li->signe = -2 ;
		}
	}
 li->indiceMax = n ;
}

void printLii( Lii* li )
{
  long int i , n ;

 n = li->taille_mantisse ;
 printf("\n Precision = %ld" , n ) ;
 printf("\n Signe = %ld\n" , li->signe ) ;
 printf(" indiceMax = %ld\n" , li->indiceMax ) ; 
 for( i = n - 1 ; i >= 0 ; i-- ) printf("%lu " , li->mantisse[i] ) ;
 printf("\n") ;
}

/*
void randomLii( Lii* li , long int positive )
// I suppose that lii is already created
// if positive == 1 then the integer is positive
// if positive == 0 then the the signis anywhere
{
 long int i , n ;
 long int temp ;

 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ )
	{
	 temp = rand() ;
	 while( temp > BASE32) temp -= BASE32 ;
	 li->mantisse[i] = temp ;
	}

 n-- ;
 while ((n>=0) && ( li->mantisse[n] == 0 )) n-- ;
 if (n<0)
	{
	 li->indiceMax = -2 ;
	 li->signe = 0 ;
	}
	else
	{
	 li->indiceMax = n ;
	 if (positive == 1) li->signe = 1 ;
	 	else
		{
		 if (positive == 0) li->signe = (rand() % 2)*2-1 ;
			else li->signe = -2 ;
		}
	}
}
*/

//--------------------------------------------------------------------
// -------   Entier long en base 2^64  avec indiceMax     ------------
//--------------------------------------------------------------------



unsigned long int initialise_fixedPoint_bini( Lii* a, unsigned long int e )
// initialise le nombre dcimal a = e.0000...
{
 long int n,signe,i ;

 n = a->taille_mantisse ;
 for(i=0;i<n;i++) a->mantisse[i] = 0 ; 
 if (n<=0) return 1 ; // The vector is not created
 if( e == 0 )
	{
	 a->signe = 0 ;
	 a->indiceMax = -1 ;
	 return 0 ;
	}
 if (e<0)
	{
	 signe = -1 ;
	 e = -e ;
	}
	else signe = 1 ;

 a->signe = signe ;
 a->mantisse[n-1] = e ;
 a->indiceMax = n-1 ;
 return 0 ;
}

void initialise_bini( Lii* a, unsigned long int e )
{
 long int n,i ;

 n = a->taille_mantisse ;
 for(i=0;i<n;i++) a->mantisse[i] = 0 ;
 if( e == 0 )
	{
	 a->signe = 0 ;
	 a->indiceMax = -1 ;
	 return ;
	}
 if ( e > 0 )
	{
	 a->signe = 1 ;
	}
	else
	{
	 a->signe = -1 ;
	 e = -e ;
	}

 a->mantisse[0] = e ;
 a->indiceMax = 0 ;
}

/*
long add_R_2P_bini_spe( Lii* a , Lii* b )
// a <- a + b
// I suppose that a > b >= 0 .
// return the overflow digit !!!
{
 long int carry = 0 ;
 long int i ;
 long int aux ;
 long int iMaxA = a->indiceMax ;
 long int iMaxB = b->indiceMax ;

 if ( b->signe ==  0 ) return 0 ;
 for( i = 0 ; i <= iMaxB ; i++ )
	{
	 aux = a->mantisse[i] + b->mantisse[i] + carry ;
	 carry = aux >> DECALAGE64_BIN ;
	 a->mantisse[i] = aux & CACHE64_BIN ;
	}
 if ( carry == 0 ) return 0 ;
 for( i = iMaxB+1 ; i <= iMaxA ; i++ )
	{
	 aux = a->mantisse[i] + carry ;
	 carry = aux >> DECALAGE64_BIN ;
	 a->mantisse[i] = aux & CACHE64_BIN ;
	 if ( carry == 0 ) break ;
	}
 if ( iMaxA < (a->taille_mantisse - 1))
	{
	 if ( carry == 1 )
		{
		 iMaxA++ ;
		 a->mantisse[iMaxA] += carry ;
		 a->indiceMax = iMaxA ;
		 return 0 ;
		}
		else return 0 ;
	}
	else return 1 ;
}

void sub_R_2P_bini_spe( Lii* a , Lii* b )
// a <- a - b
// Je suppose que a >= b >= 0 .
{
 long int carry = 0 ;
 long int i ;
 long int aux ;
 long int iMaxA = a->indiceMax ;
 long int iMaxB = b->indiceMax ;

 if ( b->signe ==  0 ) return ;
 for( i = 0 ; i <= iMaxB ; i++ )
	{
	 aux = a->mantisse[i] - b->mantisse[i] + carry ;
	 carry = aux >> 31 ;
	 a->mantisse[i] = aux - carry * BASE64_BIN ;
	}

 for( i = iMaxB+1 ; i <= iMaxA ; i++ )
	{
	 aux = a->mantisse[i] + carry ;
	 carry = aux >> 31 ;
	 a->mantisse[i] = aux - carry * BASE64_BIN ;
	 if ( carry == 0 ) break ;
	}
 while (( iMaxA >=0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if ( iMaxA < 0 ) a->signe = 0 ;
}


unsigned long int divHalfB_R_x86_64_gcc_bini( Lii* a , unsigned long int D )
// Je suppose que 0 < D < BASE32_BIN .
{
 long int i ;
 long int iMaxA = a->indiceMax ;
 long int r  ;
 unsigned long int q , temp , res ;

 if ( a->signe == 0 ) return 0 ;
 r = 0 ;

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 temp >>= HALF_DECALAGE64_BIN ;
	 r <<= HALF_DECALAGE64_BIN ;
	 temp += r ;
	 q = temp / D ;
	 r = temp - q * D ;
	 res = q << HALF_DECALAGE64_BIN ;
	 
	 temp = a->mantisse[i] ;
	 temp &= HALF_CACHE64_BIN ;
	 r <<= HALF_DECALAGE64_BIN ;
	 temp += r ;
	 q = temp / D ;
	 r = temp - q * D ;
	 res += q ;
	 //r = mulB_add_div_ul( &q,r,temp,D ) ;
	 a->mantisse[i] = res ;
	}

 while (( iMaxA >= 0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( a->mantisse[0] == 0 )) a->signe = 0 ;
 	else a->signe = 1 ;

 return r ;
}
*/



/*
long int divB_R_bini( Lii* a , long int D )
// Je suppose que 0 < D < BASE32_BIN .
{
 long int i ;
 long int iMaxA = a->indiceMax ;
 long int r  ;
 unsigned long int q , temp ;

 if ( a->signe == 0 ) return 0 ;
 r = 0 ;

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 //r = mulB_add_div_ul( &q,r,temp,D ) ;
	 a->mantisse[i] = q ;
	}

 while (( iMaxA >= 0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( a->mantisse[0] == 0 )) a->signe = 0 ;
 	else a->signe = 1 ;

 return r ;
}*/

/*
long int divB_bini( Lii* res , Lii* a , long int D )
// Je suppose que 0 < D < BASE32_BIN .
// res <- a / D
{
 long int i , r=0 ;
 long int iMaxA = a->indiceMax ;
 unsigned long int temp , q ;

 if ( a->signe == 0 )
	{
	 res->signe = 0 ;
	 return 0 ;
	}

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 //r = mulB_add_div_ul( &q,r,temp,D ) ;
	 res->mantisse[i] = q ;
	}

 while (( iMaxA >= 0 ) && ( res->mantisse[iMaxA] == 0 )) iMaxA-- ;
 res->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( res->mantisse[0] == 0 ))
 	res->signe = 0 ; else res->signe = 1 ;

 return r ;
}
*/

/*
long mulB_R_bini( Lii* a , long m )
// Cette routine calcule le produit a <- m * a o
// 0 <= m < BASE32_BIN .
{
 long i ;
 long iMaxA = a->indiceMax ;
 unsigned long temp, carry=0 ;

 if ( a->signe == 0 ) return 0 ;
 if ( m == 0 )
	{
	 a->signe = 0 ;
	 return 0 ;
	}
 for( i = 0 ; i <= iMaxA ; i++ )
	{
	 temp = a->mantisse[i] ;
	 //a->mantisse[i] = mul_add_div_normalized_bin( &carry,m,temp,carry ) ;
	}


 if ( carry != 0 )
 	{
	 if ( (iMaxA + 1) == a->taille_mantisse ) return carry ;
		else
		{
		 a->indiceMax = iMaxA + 1 ;
		 a->mantisse[iMaxA+1] = carry ;
		 return 0 ;
		}
	}
	else return 0 ;
}
*/

void print_cst_base2_bini( Lii* a )
// affiche e , pi , sqrt(2) ... en base 2.
{
 long int i = 1 ;
 long int j , exposant ;
 unsigned long int temp, temp2 ;
 long int n = a->taille_mantisse - 1 ;


 printf( "\n" ) ;
 exposant  = 0 ;
 temp  = a->mantisse[n] ;
 temp2 = temp ;
 for( exposant = 0 ; temp2 != 0 ; exposant++ ) temp2 >>= 1 ;
 exposant-- ;
 while( exposant >= 0 )
   {
     printf("%lu" , ((temp >> exposant) & PARITE)  ) ;
     exposant-- ;
   }

 printf( ".\n" ) ;
 j = n - 1 ;
 i = 1 ;
 while( j >= 0 )
   {
     temp  = a->mantisse[j] ;
     for( exposant = DECALAGE64_BIN - 1 ; exposant >= 0 ; exposant-- )
       {
	 printf("%lu" , ((temp >> exposant) & PARITE ) ) ;
	 if ( (i % 100) == 0 ) printf("  :: %ld \n" , i ) ;
         if (( (i % 10) == 0 ) && ( (i % 100) != 0 )) printf(" ") ;
	 i++ ;
       }
     j-- ;
   }
 printf("\n\n") ;
}

void print_cst_base16_bini( Lii* a )
// affiche e , pi , sqrt(2) ... en base 16.
{
 long int i = 1 ;
 long int j , exposant ;
 unsigned long int temp, temp2, chiffre ;
 long int n = a->taille_mantisse - 1 ;


 
 printf( "\n" ) ;
 exposant  = 0 ;
 temp  = a->mantisse[n] ;
 temp2 = temp ;
 for( exposant = 0 ; temp2 != 0 ; exposant++ ) temp2 >>= 4 ;
 exposant-- ;
 while( exposant >= 0 )
   {
     printf("%lu" , ((temp >> ( exposant << 4 )) & CACHE_BASE16 )  ) ;
     exposant-- ;
   }

 printf( ".\n" ) ;
 j = n - 1 ;
 i = 1 ;
 while( j >= 0 )
   {
	temp  = a->mantisse[j] ;
	for( exposant = DECALAGE64_BIN - 4 ; exposant >= 0 ; exposant -= 4 )
         	{
		 chiffre = (temp >> exposant )  & CACHE_BASE16 ;
		 //printf("%i\n" , chiffre ) ;
		 if ( chiffre < 10 ) printf("%i" , (int) chiffre ) ;
		 	        else printf("%c" , (char) chiffre + 55 ) ;
		 if ( (i % 100) == 0 ) printf("  :: %ld \n" , i ) ;
        	 if (( (i % 10) == 0 ) && ( (i % 100) != 0 )) printf(" ") ;
		 i++ ;
		}
	j-- ;
   }
 printf("\n\n") ;
}




//-----------------------------------------------------------------------------
// -------------   Entier long en base 2^30  de bas niveau     ----------------
//-----------------------------------------------------------------------------

void mul_2P_naif_gcc_bin( long int* prod , long int* a , long int* b , long int n )
{
 unsigned long int* tempA ;
 unsigned long int* tempB ;
 unsigned long int* tempP ;
 long int n_2,i,j;
 unsigned long int m,carry,tempon ;
 
 n_2 = n << 1 ;
 tempA = (unsigned long int*) calloc( n_2 , sizeof(long int) ) ;
 tempB = (unsigned long int*) calloc( n_2 , sizeof(long int) ) ; 
 tempP = (unsigned long int*) calloc( n_2*2 , sizeof(long int) ) ; 
 for(i=0;i<n_2*2;i++) tempP[i] = 0 ;
 for(i=0;i<n;i++)
	{
	 tempA[2*i] = a[i] & HALF_CACHE64_BIN ;
	 tempA[2*i+1] = a[i] >>  HALF_DECALAGE64_BIN ; 
	}
 for(i=0;i<n;i++)
	{
	 tempB[2*i] = b[i] & HALF_CACHE64_BIN ;
	 tempB[2*i+1] = b[i] >>  HALF_DECALAGE64_BIN ; 
	} 
 for( i = 0 ; i < n_2 ; i++ )
        {
         m = tempB[i] ;
         carry = 0 ;
         for( j = 0 ; j < n_2 ; j++ )
                {
		 tempon  = tempA[j] * m ;
		 tempon += carry ;
		 tempon += tempP[i+j] ;
		 carry = tempon >> HALF_DECALAGE64_BIN ;
		 tempP[i+j] = tempon & HALF_CACHE64_BIN ;
                }
         tempP[i+n_2] += carry ;
        }
 for(i=0;i<n_2;i++) prod[i] = (tempP[2*i+1] << HALF_DECALAGE64_BIN) + tempP[2*i] ;
 free(tempA);
 free(tempB); 
 free(tempP);
} 





