/*
 * Routine conue par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef LFX86_64_H
#define LFX86_64_H


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
//#include "op_fftp.h"

#define DECALAGE64_BIN 64
#define HALF_CACHE64_BIN 0x00000000ffffffff
#define HALF_DECALAGE64_BIN 32
#define CACHE_BASE16 15L
#define PARITE 1L


// Li reprsente un entier long 
// Le tableau mantisse stocke les chiffres dans la base 2^64
// mantisse[0] contient le chiffre des units, etc...

typedef struct
    {
      long int  signe ;
      long int taille_mantisse ;
      unsigned long int* mantisse ;
      long int aux ;
      
} Li ;

// idem que pour Li
// La seule difffrence est que tous les calculs se font sur les
// indicemMax  premiers chiffres en partant de mantisse[0]. 
typedef struct
    {
     long int signe;
     long int taille_mantisse ; 
     unsigned long int* mantisse ;
     long int aux ;
     long int indiceMax ;
} Lii ;


unsigned long int createLi( Li* li , long int nb ) ;
void destroyLi( Li* li ) ;
long int indiceMax( long int* chiffre , Li* a ) ;
void printLi( Li* li ) ;
void copy( Li* nouveau , Li* li ) ;
void copyAndCreate( Li* nouveau , Li* li ) ;
void randomLi( Li* li , long positive ) ;
//void permutLi( Li* li2 , Li* li ) ; // je suppose que li et li2 ont mme taille

//-----------------------------------------------------------------------------
// ----------------   Entier long en base 2^64      ---------------------------
//-----------------------------------------------------------------------------

	void add_R_2P_bin( Li* a , Li* b ) ;
extern 	void add_2P_x86_64_bin( Li* result , Li* a , Li* b ) ;
	void sub_2P_bin( Li* result , Li* a , Li* b ) ;
	void sub_R_2P_bin( Li* a , Li* b ) ;
	void add_bin( Li* result , Li* a , Li* b ) ;
	void add_R_bin( Li* a , Li* b ) ;
	void sub_bin( Li* result , Li* a , Li* b ) ;
	void sub_R_bin( Li* a , Li* b ) ;
	void mulB_bin( Li* prod , Li* li , long m ) ;// suppose que prod est dj cr
	unsigned long int div_by_pow2_bin( Li* result , Li* li , long p ) ;// suppose que result est dj cr
	void mul_bin( Li* prod , Li* a , Li* b ) ;// suppose que prod, result sont dj crs
	unsigned long int mul_trunc_bin( Li* prod , Li* a , Li* b ) ;
	unsigned long int square_trunc_bin( Li* prod , Li* a ) ;
	void square_bin( Li* prod , Li* a ) ;
	void mul_by2_R_bin( Li* a ) ;
	double BigFloatToFloat( Li* bfBaseBin ) ;

//-----------------------------------------------------------------------------
void createLii( Lii* li , long nb ) ;
void destroyLii( Lii* li ) ;
void copyLii(  Lii* nouveau ,Lii* li ) ;
void printLii( Lii* li ) ;
void randomLii( Lii* li,long int positive ) ;
void printLii( Lii* li ) ;
//void randomLii(Lii* li, unsigned long period ) ;


//-----------------------------------------------------------------------------
// -------------   Entier long en base 2^64  avec indiceMax     ---------------
//-----------------------------------------------------------------------------

       unsigned long int initialise_fixedPoint_bini( Lii* le , unsigned long int e ) ;
       void initialise_bini( Lii* le , unsigned long int e ) ;
extern unsigned long int add_R_2P_x86_64_bini_spe( Lii* a , Lii* b ) ;
extern void sub_R_2P_x86_64_bini_spe( Lii* a , Lii* b ) ;
extern unsigned long int divHalfB_R_x86_64_bini( Lii* a , long int D ) ;
extern unsigned long int divB_R_x86_64_bini( Lii* a , unsigned long int D ) ;
extern unsigned long int divB_x86_64_bini( Lii* res , Lii* a , unsigned long int D ) ;
extern unsigned long int divHalfB_x86_64_bini( Lii* res , Lii* a , unsigned long int D ) ;
extern unsigned long int mulB_R_x86_64_bini( Lii* a , unsigned long int m ) ;
       void print_cst_base2_bini( Lii* a ) ;
       void print_cst_base16_bini( Lii* a ) ;

//-----------------------------------------------------------------------------
// -------------   Entier long en base 2^64  de bas niveau     ----------------
//-----------------------------------------------------------------------------

extern	void mul_2P_naif_x86_64_bin( long int* prod , long int* a , long int* b , long int n ) ;
	void mul_2P_naif_gcc_bin( long int* prod , long int* a , long int* b , long int n ) ;
extern	unsigned long int square_2P_naif_x86_64_bin( long* prod , long* a , long n ) ;
//void square_2P_naif_bin( long* prod , long* a , long n ) ;





#endif

